@@ -31,21 +31,23 @@
 #include "../cqp/corpmanag.h"
 #include "../cqp/groups.h"
 
+void Rprintf(const char *, ...);
 
-char *user = "", *passwd = "";
+/** String containing the username sent by the currently-connect CQi client */
+char *user = "";
+/** String containing the password sent by the currently-connect CQi client */
+char *passwd = "";
 
 
-/*
- *
- *  CQi server welcome & copyright message
- *
+/**
+ *  Prints the CQi server welcome and copyright message.
  */
-
 void 
-cqiserver_welcome(void) {
-  printf("** CQPserver v" VERSION "\n");
-  printf("** implementing version %d.%d of the CQi\n", CQI_MAJOR_VERSION, CQI_MINOR_VERSION);
-  printf("\n");
+cqiserver_welcome(void)
+{
+  Rprintf("** CQPserver v" VERSION "\n");
+  Rprintf("** implementing version %d.%d of the CQi\n", CQI_MAJOR_VERSION, CQI_MINOR_VERSION);
+  Rprintf("\n");
 }
 
 
@@ -55,22 +57,44 @@ cqiserver_welcome(void) {
  *
  */
 
+/**
+ * Shuts down the server with an "unknown CQi command" error condition.
+ *
+ * @param cmd  The integer representing the unknown command received from the client.
+ */
 void 
-unknown_command_error(int cmd) {
-  fprintf(stderr, "CQPserver: unknown CQi command 0x%04X.\n", cmd);
+cqiserver_unknown_command_error(int cmd)
+{
+  Rprintf("CQPserver: unknown CQi command 0x%04X.\n", cmd);
   exit(1);
 }
 
+/**
+ * Shuts down the server with an "CQi command not allowed here" error condition.
+ *
+ * @param cmd  The integer representing the wrong command received from the client.
+ */
 void 
-wrong_command_error(int cmd) {
-  fprintf(stderr, "CQPserver: command 0x%04X not allowed in this context.\n", cmd);
+cqiserver_wrong_command_error(int cmd)
+{
+  Rprintf("CQPserver: command 0x%04X not allowed in this context.\n", cmd);
   exit(1);
 }
 
+/**
+ * Shuts down the server with an "internal error" condition.
+ *
+ * Both parameters will be printed as part of the shutdown error message.
+ *
+ * @param function  String: should be name of the calling function, that is,
+ *                  the point where the error was raised.
+ * @param reason    String containing any other explanatory details about the error.
+ */
 void
-internal_error(char *function, char *reason) {
-  fprintf(stderr, "CQPserver: internal error in %s()\n", function);
-  fprintf(stderr, "CQPserver: ''%s''\n", reason);
+cqiserver_internal_error(char *function, char *reason)
+{
+  Rprintf("CQPserver: internal error in %s()\n", function);
+  Rprintf("CQPserver: ''%s''\n", reason);
   exit(1);
 }
 
@@ -81,12 +105,26 @@ internal_error(char *function, char *reason) {
  *
  */
 
-/* send CQI_CL_ERROR_* according to the CL cderrno variable */
+/**
+ * Sends the current CL error value to the client.
+ *
+ * This function takes the current contents of of the CL library's global
+ * cl_errno error value and sends it to the client.
+ *
+ * It takes the CL error consant and translates it into the corresponding
+ * CQI_CL_ERROR_* constant.
+ *
+ * NB: This function shuts down the server with an error condition if cl_errno
+ * does not actually contain an error condition.
+ *
+ * @see cl_errno
+ */
 void
-send_cl_error(void) {
+send_cl_error(void)
+{
   int cmd;
   
-  switch (cderrno) {
+  switch (cl_errno) {
   case CDA_EATTTYPE:
     cmd = CQI_CL_ERROR_WRONG_ATTRIBUTE_TYPE;
     break;
@@ -110,14 +148,14 @@ send_cl_error(void) {
     cmd = CQI_CL_ERROR_INTERNAL;
     break;
   case CDA_OK:
-    fprintf(stderr, "CQPserver: send_cl_error() called with cderrno == CDA_OK\n");
+    Rprintf("CQPserver: send_cl_error() called with cderrno == CDA_OK\n");
     exit(1);
   default:
-    fprintf(stderr, "CQPserver: send_cl_error() unknown value in cderrno\n");
+    Rprintf("CQPserver: send_cl_error() unknown value in cderrno\n");
     exit(1);
   }
   if (server_debug)
-    fprintf(stderr, "CQi: CL error, returning 0x%04X\n", cmd);
+    Rprintf("CQi: CL error, returning 0x%04X\n", cmd);
   cqi_command(cmd);
   return;
 }
@@ -125,17 +163,18 @@ send_cl_error(void) {
 
 /*
  *
- *  CQi commands  (removed from interpreter loop)
+ *  CQi commands  (called from interpreter loop)
  *
  */
 
 void
-do_cqi_corpus_list_corpora(void) {
+do_cqi_corpus_list_corpora(void)
+{
   CorpusList *cl;
   int n = 0;
   
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CORPUS_LIST_CORPORA()\n");
+    Rprintf("CQi: CQI_CORPUS_LIST_CORPORA()\n");
   /* ugly, but it's easiest ... first count corpora, then return names one by one */
   for (cl = FirstCorpusFromList(); cl != NULL; cl = NextCorpusFromList(cl)) {
     if (cl->type == SYSTEM)
@@ -151,24 +190,32 @@ do_cqi_corpus_list_corpora(void) {
 }
 
 void 
-do_cqi_corpus_charset(void) {
+do_cqi_corpus_charset(void)
+{
   char *c;
-
+  CorpusList *cl;
   c = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CORPUS_CHARSET('%s')\n", c);
-  /* this is a dummy until we've implemented the registry extensions */
-  cqi_data_string("latin1");
+    Rprintf("CQi: CQI_CORPUS_CHARSET('%s')\n", c);
+
+  cl = findcorpus(c, SYSTEM, 0);
+  if (cl == NULL || !access_corpus(cl)) {
+    cqi_command(CQI_CQP_ERROR_NO_SUCH_CORPUS);
+  }
+  else {
+    cqi_data_string(cl_charset_name(cl->corpus->charset));
+  }
   free(c);
 }
 
 void 
-do_cqi_corpus_properties(void) {
+do_cqi_corpus_properties(void)
+{
   char *c;
 
   c = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CORPUS_PROPERTIES('%s')\n", c);
+    Rprintf("CQi: CQI_CORPUS_PROPERTIES('%s')\n", c);
   /* this is a dummy until we've implemented the registry extensions */
   cqi_data_string_list(NULL, 0);
   free(c);
@@ -176,7 +223,8 @@ do_cqi_corpus_properties(void) {
 
 /* this part sends attributes of a certain type as a STRING[] to the client */
 void
-send_cqi_corpus_attributes(Corpus *c, int type) {
+send_cqi_corpus_attributes(Corpus *c, int type)
+{
   Attribute *a;
   int len;
   
@@ -194,7 +242,8 @@ send_cqi_corpus_attributes(Corpus *c, int type) {
 }
 
 void
-do_cqi_corpus_attributes(int type) {
+do_cqi_corpus_attributes(int type)
+{
   char *c, *typename;
   CorpusList *cl;
 
@@ -214,7 +263,7 @@ do_cqi_corpus_attributes(int type) {
       cqi_general_error("INTERNAL ERROR: do_cqi_corpus_attributes(): unknown attribute type");
       return;
     }
-    fprintf(stderr, "CQi: CQI_CORPUS_%s_ATTRIBUTES('%s')\n", typename, c);
+    Rprintf("CQi: CQI_CORPUS_%s_ATTRIBUTES('%s')\n", typename, c);
   }
   cl = findcorpus(c, SYSTEM, 0);
   if (cl == NULL || !access_corpus(cl)) {
@@ -227,13 +276,14 @@ do_cqi_corpus_attributes(int type) {
 }
 
 void
-do_cqi_corpus_full_name(void) {
+do_cqi_corpus_full_name(void)
+{
   char *c;
   CorpusList *cl;
 
   c = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CORPUS_FULL_NAME('%s')\n", c);
+    Rprintf("CQi: CQI_CORPUS_FULL_NAME('%s')\n", c);
   cl = findcorpus(c, SYSTEM, 0);
   if (cl == NULL || !access_corpus(cl)) {
     cqi_command(CQI_CQP_ERROR_NO_SUCH_CORPUS);
@@ -251,7 +301,7 @@ do_cqi_corpus_structural_attribute_has_values(void) {
           
   a = cqi_read_string();        /* need to try all possible attribute types */
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES('%s')\n", a);
+    Rprintf("CQi: CQI_CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES('%s')\n", a);
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute != NULL) {
     cqi_data_bool(cl_struc_values(attribute));
@@ -263,14 +313,15 @@ do_cqi_corpus_structural_attribute_has_values(void) {
 }
 
 void
-do_cqi_cl_attribute_size(void) {
+do_cqi_cl_attribute_size(void)
+{
   char *a;
   Attribute *attribute;
   int size;
           
   a = cqi_read_string();        /* need to try all possible attribute types */
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_ATTRIBUTE_SIZE('%s')\n", a);
+    Rprintf("CQi: CQI_CL_ATTRIBUTE_SIZE('%s')\n", a);
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute != NULL) {
     size = cl_max_cpos(attribute);
@@ -314,14 +365,15 @@ do_cqi_cl_attribute_size(void) {
 }
 
 void
-do_cqi_cl_lexicon_size(void) {
+do_cqi_cl_lexicon_size(void)
+{
   char *a;
   Attribute *attribute;
   int size;
           
   a = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_LEXICON_SIZE('%s')\n", a);
+    Rprintf("CQi: CQI_CL_LEXICON_SIZE('%s')\n", a);
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute != NULL) {
     size = cl_max_id(attribute);
@@ -339,16 +391,18 @@ do_cqi_cl_lexicon_size(void) {
 }
 
 void
-do_cqi_cl_drop_attribute(void) {
+do_cqi_cl_drop_attribute(void)
+{
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_DROP_ATTRIBUTE()  --  not implemented\n");
+    Rprintf("CQi: CQI_CL_DROP_ATTRIBUTE()  --  not implemented\n");
  cqi_general_error("CQI_CL_DROP_ATTRIBUTE not implemented."); 
 }
 
 /* one might wish to add extensive error checking to all the CL functions,
    but that will need a LOT of code! */
 void
-do_cqi_cl_str2id(void) {
+do_cqi_cl_str2id(void)
+{
   char **strlist;
   int len, i, id;
   char *a;
@@ -357,10 +411,10 @@ do_cqi_cl_str2id(void) {
   a = cqi_read_string();
   len = cqi_read_string_list(&strlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_STR2ID('%s', [", a);
+    Rprintf("CQi: CQI_CL_STR2ID('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "'%s' ", strlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("'%s' ", strlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -385,7 +439,8 @@ do_cqi_cl_str2id(void) {
 }
 
 void
-do_cqi_cl_id2str(void) {
+do_cqi_cl_id2str(void)
+{
   int *idlist;
   int len, i;
   char *a, *str;
@@ -394,10 +449,10 @@ do_cqi_cl_id2str(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&idlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_ID2STR('%s', [", a);
+    Rprintf("CQi: CQI_CL_ID2STR('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", idlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", idlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -420,7 +475,8 @@ do_cqi_cl_id2str(void) {
 }
 
 void
-do_cqi_cl_id2freq(void) {
+do_cqi_cl_id2freq(void)
+{
   int *idlist;
   int len, i, f;
   char *a;
@@ -429,10 +485,10 @@ do_cqi_cl_id2freq(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&idlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_ID2FREQ('%s', [", a);
+    Rprintf("CQi: CQI_CL_ID2FREQ('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", idlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", idlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -457,7 +513,8 @@ do_cqi_cl_id2freq(void) {
 }
 
 void
-do_cqi_cl_cpos2str(void) {
+do_cqi_cl_cpos2str(void)
+{
   int *cposlist;
   int len, i;
   char *a, *str;
@@ -466,10 +523,10 @@ do_cqi_cl_cpos2str(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2STR('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2STR('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -492,7 +549,8 @@ do_cqi_cl_cpos2str(void) {
 }
 
 void
-do_cqi_cl_cpos2id(void) {
+do_cqi_cl_cpos2id(void)
+{
   int *cposlist;
   int len, i, id;
   char *a;
@@ -501,10 +559,10 @@ do_cqi_cl_cpos2id(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2ID('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2ID('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -529,7 +587,8 @@ do_cqi_cl_cpos2id(void) {
 }
 
 void
-do_cqi_cl_cpos2struc(void) {
+do_cqi_cl_cpos2struc(void)
+{
   int *cposlist;
   int len, i, struc;
   char *a;
@@ -538,10 +597,10 @@ do_cqi_cl_cpos2struc(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2STRUC('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2STRUC('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute == NULL) {
@@ -569,7 +628,8 @@ do_cqi_cl_cpos2struc(void) {
    and should be improved with a caching model to avoid the frequent cpos2struc lookup;
    perhaps make them CL functions with an intelligent caching algorithm? */
 void
-do_cqi_cl_cpos2lbound(void) {
+do_cqi_cl_cpos2lbound(void)
+{
   int *cposlist;
   int len, i, struc, lb, rb;
   char *a;
@@ -578,10 +638,10 @@ do_cqi_cl_cpos2lbound(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2LBOUND('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2LBOUND('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute == NULL) {
@@ -611,7 +671,8 @@ do_cqi_cl_cpos2lbound(void) {
 }
 
 void
-do_cqi_cl_cpos2rbound(void) {
+do_cqi_cl_cpos2rbound(void)
+{
   int *cposlist;
   int len, i, struc, lb, rb;
   char *a;
@@ -620,10 +681,10 @@ do_cqi_cl_cpos2rbound(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2RBOUND('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2RBOUND('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute == NULL) {
@@ -653,7 +714,8 @@ do_cqi_cl_cpos2rbound(void) {
 }
 
 void
-do_cqi_cl_cpos2alg(void) {
+do_cqi_cl_cpos2alg(void)
+{
   int *cposlist;
   int len, i, alg;
   char *a;
@@ -662,10 +724,10 @@ do_cqi_cl_cpos2alg(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&cposlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_CPOS2ALG('%s', [", a);
+    Rprintf("CQi: CQI_CL_CPOS2ALG('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", cposlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", cposlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_ALIGN);
   if (attribute == NULL) {
@@ -690,7 +752,8 @@ do_cqi_cl_cpos2alg(void) {
 }
 
 void
-do_cqi_cl_struc2str(void) {
+do_cqi_cl_struc2str(void)
+{
   int *struclist;
   int len, i;
   char *a, *str;
@@ -699,10 +762,10 @@ do_cqi_cl_struc2str(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&struclist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_STRUC2STR('%s', [", a);
+    Rprintf("CQi: CQI_CL_STRUC2STR('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", struclist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", struclist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute == NULL) {
@@ -725,7 +788,8 @@ do_cqi_cl_struc2str(void) {
 }
 
 void
-do_cqi_cl_id2cpos(void) {
+do_cqi_cl_id2cpos(void)
+{
   int *cposlist;
   int len, id;
   char *a;
@@ -734,7 +798,7 @@ do_cqi_cl_id2cpos(void) {
   a = cqi_read_string();
   id = cqi_read_int();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_ID2CPOS('%s', %d)\n", a, id);
+    Rprintf("CQi: CQI_CL_ID2CPOS('%s', %d)\n", a, id);
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
     cqi_command(cqi_errno);
@@ -752,7 +816,8 @@ do_cqi_cl_id2cpos(void) {
 }
 
 void
-do_cqi_cl_idlist2cpos(void) {
+do_cqi_cl_idlist2cpos(void)
+{
   int *idlist, *cposlist;
   int i, len, cposlen;
   char *a;
@@ -761,10 +826,10 @@ do_cqi_cl_idlist2cpos(void) {
   a = cqi_read_string();
   len = cqi_read_int_list(&idlist);
   if (server_debug) {
-    fprintf(stderr, "CQi: CQI_CL_IDLIST2CPOS('%s', [", a);
+    Rprintf("CQi: CQI_CL_IDLIST2CPOS('%s', [", a);
     for (i=0; i<len; i++)
-      fprintf(stderr, "%d ", idlist[i]);
-    fprintf(stderr, "])\n");
+      Rprintf("%d ", idlist[i]);
+    Rprintf("])\n");
   }
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
@@ -786,7 +851,8 @@ do_cqi_cl_idlist2cpos(void) {
 }
 
 void
-do_cqi_cl_regex2id(void) {
+do_cqi_cl_regex2id(void)
+{
   int *idlist;
   int len;
   char *a, *regex;
@@ -795,7 +861,7 @@ do_cqi_cl_regex2id(void) {
   a = cqi_read_string();
   regex = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_REGEX2ID('%s', '%s')\n", a, regex);
+    Rprintf("CQi: CQI_CL_REGEX2ID('%s', '%s')\n", a, regex);
   attribute = cqi_lookup_attribute(a, ATT_POS);
   if (attribute == NULL) {
     cqi_command(cqi_errno);
@@ -818,7 +884,8 @@ do_cqi_cl_regex2id(void) {
 }
 
 void
-do_cqi_cl_struc2cpos(void) {
+do_cqi_cl_struc2cpos(void)
+{
   int struc, start, end;
   char *a;
   Attribute *attribute;
@@ -826,7 +893,7 @@ do_cqi_cl_struc2cpos(void) {
   a = cqi_read_string();
   struc = cqi_read_int();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_STRUC2CPOS('%s', %d)\n", a, struc);
+    Rprintf("CQi: CQI_CL_STRUC2CPOS('%s', %d)\n", a, struc);
   attribute = cqi_lookup_attribute(a, ATT_STRUC);
   if (attribute == NULL) {
     cqi_command(cqi_errno);
@@ -841,7 +908,8 @@ do_cqi_cl_struc2cpos(void) {
 }
 
 void
-do_cqi_cl_alg2cpos(void) {
+do_cqi_cl_alg2cpos(void)
+{
   int alg, s1, s2, t1, t2;
   char *a;
   Attribute *attribute;
@@ -849,7 +917,7 @@ do_cqi_cl_alg2cpos(void) {
   a = cqi_read_string();
   alg = cqi_read_int();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CL_ALG2CPOS('%s', %d)\n", a, alg);
+    Rprintf("CQi: CQI_CL_ALG2CPOS('%s', %d)\n", a, alg);
   attribute = cqi_lookup_attribute(a, ATT_ALIGN);
   if (attribute == NULL) {
     cqi_command(cqi_errno);
@@ -864,14 +932,15 @@ do_cqi_cl_alg2cpos(void) {
 }
 
 void
-do_cqi_cqp_list_subcorpora(void) {
+do_cqi_cqp_list_subcorpora(void)
+{
   char *corpus;
   CorpusList *cl, *mother;
   int n = 0;
   
   corpus = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CQP_LIST_SUBCORPORA(%s)\n", corpus);
+    Rprintf("CQi: CQI_CQP_LIST_SUBCORPORA(%s)\n", corpus);
   mother = cqi_find_corpus(corpus);
   if (!check_corpus_name(corpus) || mother == NULL) 
     cqi_command(cqi_errno);
@@ -894,30 +963,39 @@ do_cqi_cqp_list_subcorpora(void) {
   free(corpus);
 }
 
-/* CQP queries must be terminated with a single semicolon; */
-/* multiple semicolons will produce an error to occur -- so we */
-/* have to check and add a semicolon if necessary. */
+/**
+ * Tests whether or nto the final non-blank character in a string is a semicolon.
+ *
+ * CQP queries must be terminated with a single semicolon;
+ * multiple semicolons will produce an error to occur -- so we
+ * have to check and add a semicolon if necessary.
+ *
+ * @return  Boolean: true iff the final non-blank character is a semicolon.
+ */
 int
-query_has_semicolon(char *query) {
+query_has_semicolon(char *query)
+{
   char *p;
 
   if (query == NULL || *query == 0)
     return 0;
   p = query + strlen(query); 
   while (--p > query)           /* stop at first non-blank char or at first string character */
-    if (!(*p == ' ' || *p == '\t')) break;
+    if (!(*p == ' ' || *p == '\t'))
+      break;
   return (*p == ';') ? 1 : 0;
 }
 
 void 
-do_cqi_cqp_query(void) {
+do_cqi_cqp_query(void)
+{
   char *child, *mother, *query, *c, *sc;
   
   mother = cqi_read_string();
   child = cqi_read_string();
   query = cqi_read_string();
   if (server_debug) 
-    fprintf(stderr, "CQi: CQI_CQP_QUERY('%s', '%s', '%s')\n", mother, child, query);
+    Rprintf("CQi: CQI_CQP_QUERY('%s', '%s', '%s')\n", mother, child, query);
   if (!split_subcorpus_spec(mother, &c, &sc)) {
     cqi_command(cqi_errno);
   }
@@ -932,7 +1010,7 @@ do_cqi_cqp_query(void) {
     else {
       query_lock = floor(1e9 * cl_runif()) + 1; /* activate query lock mode with random key */
 
-      printf("CQPSERVER: query_lock = %d\n", query_lock);
+      Rprintf("CQPSERVER: query_lock = %d\n", query_lock);
       if (query_has_semicolon(query))
         sprintf(cqp_query, "%s = %s", child, query);
       else
@@ -949,9 +1027,9 @@ do_cqi_cqp_query(void) {
           cqi_command(CQI_CQP_ERROR_GENERAL);
         else {
           if (server_log) {
-            printf("'%s' ran the following query on %s\n", user, mother);
-            printf("\t%s\n", cqp_query);
-            printf("and got %d matches.\n", childcl->size);
+            Rprintf("'%s' ran the following query on %s\n", user, mother);
+            Rprintf("\t%s\n", cqp_query);
+            Rprintf("and got %d matches.\n", childcl->size);
           }
           cqi_command(CQI_STATUS_OK);
 
@@ -968,13 +1046,14 @@ do_cqi_cqp_query(void) {
 }
 
 void
-do_cqi_cqp_subcorpus_size(void) {
+do_cqi_cqp_subcorpus_size(void)
+{
   char *subcorpus;
   CorpusList *cl;
 
   subcorpus = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CQP_SUBCORPUS_SIZE('%s')\n", subcorpus);
+    Rprintf("CQi: CQI_CQP_SUBCORPUS_SIZE('%s')\n", subcorpus);
   cl = cqi_find_corpus(subcorpus);
   if (cl == NULL) 
     cqi_command(cqi_errno);
@@ -984,7 +1063,13 @@ do_cqi_cqp_subcorpus_size(void) {
   free(subcorpus);
 }
 
-/* used for debugging output & to check valid fields in subroutines below */
+/**
+ * Returns string representations of CQI_CONST_FIELD_ values.
+ *
+ * Utility function, used for debugging output & to check valid fields in subroutines below.
+ *
+ * TODO as a utiltiy, shouldn't this be in the cqi library (server.c?)
+ */
 char *
 cqi_field_name(cqi_byte field) {
   switch (field) {
@@ -1002,7 +1087,8 @@ cqi_field_name(cqi_byte field) {
 }
 
 void
-do_cqi_cqp_subcorpus_has_field(void) {
+do_cqi_cqp_subcorpus_has_field(void)
+{
   char *subcorpus;
   CorpusList *cl;
   cqi_byte field;
@@ -1018,7 +1104,7 @@ do_cqi_cqp_subcorpus_has_field(void) {
     field_ok = 0;
   }
   if (server_debug) 
-    fprintf(stderr, "CQi: CQI_CQP_SUBCORPUS_HAS_FIELD('%s', %s)\n", 
+    Rprintf("CQi: CQI_CQP_SUBCORPUS_HAS_FIELD('%s', %s)\n", 
             subcorpus, fieldname);
 
   cl = cqi_find_corpus(subcorpus);
@@ -1047,7 +1133,7 @@ do_cqi_cqp_subcorpus_has_field(void) {
         cqi_data_bool(CQI_CONST_YES);
       break;
     default:
-      internal_error("do_cqi_cqp_subcorpus_has_field", "Can't identify requested field.");
+      cqiserver_internal_error("do_cqi_cqp_subcorpus_has_field", "Can't identify requested field.");
     }
     cqi_flush();
   }
@@ -1055,15 +1141,28 @@ do_cqi_cqp_subcorpus_has_field(void) {
   free(subcorpus);
 }
 
-/* CQI_CQP_DUMP_SUBCORPUS returns list of (-1) values if requested field is not set */
+/**
+ * Sends n instances of integer -1 to the client.
+ *
+ * Utility function for do_cqi_cqp_dump_subcorpus().
+ *
+ * This is the error condition of the CQI_CQP_DUMP_SUBCORPUS command:
+ * it returns a list of (-1) values if requested field is not set.
+ *
+ * It is assumed that the length of the lsit has already been sent.
+ *
+ * @param n  Length of list to send.
+ */
 void
-cqi_send_minus_one_list(int n) {
+do_cqi_send_minus_one_list(int n)
+{
   while (n--) 
     cqi_send_int(-1);
 }
 
 void
-do_cqi_cqp_dump_subcorpus(void) {
+do_cqi_cqp_dump_subcorpus(void)
+{
   char *subcorpus;
   CorpusList *cl;
   cqi_byte field;
@@ -1082,7 +1181,7 @@ do_cqi_cqp_dump_subcorpus(void) {
     field_ok = 0;
   }
   if (server_debug) 
-    fprintf(stderr, "CQi: CQI_CQP_DUMP_SUBCORPUS('%s', %s, %d, %d)\n", 
+    Rprintf("CQi: CQI_CQP_DUMP_SUBCORPUS('%s', %s, %d, %d)\n", 
             subcorpus, fieldname, first, last);
 
   cl = cqi_find_corpus(subcorpus);
@@ -1107,20 +1206,20 @@ do_cqi_cqp_dump_subcorpus(void) {
         break;
       case CQI_CONST_FIELD_TARGET:
         if (cl->targets == NULL) 
-          cqi_send_minus_one_list(size);
+          do_cqi_send_minus_one_list(size);
         else 
           for (i=first; i<=last; i++)
             cqi_send_int(cl->targets[i]);
         break;
       case CQI_CONST_FIELD_KEYWORD:
         if (cl->keywords == NULL) 
-          cqi_send_minus_one_list(size);
+          do_cqi_send_minus_one_list(size);
         else 
           for (i=first; i<=last; i++)
             cqi_send_int(cl->keywords[i]);
         break;
       default:
-        internal_error("do_cqi_cqp_dump_subcorpus", "No handler for requested field.");
+        cqiserver_internal_error("do_cqi_cqp_dump_subcorpus", "No handler for requested field.");
       }
       cqi_flush();
   }
@@ -1129,14 +1228,15 @@ do_cqi_cqp_dump_subcorpus(void) {
 }
 
 void
-do_cqi_cqp_drop_subcorpus(void) {
+do_cqi_cqp_drop_subcorpus(void)
+{
   char *subcorpus;
   CorpusList *cl;
   char *c, *sc;
 
   subcorpus = cqi_read_string();
   if (server_debug)
-    fprintf(stderr, "CQi: CQI_CQP_DROP_SUBCORPUS('%s')\n", subcorpus);
+    Rprintf("CQi: CQI_CQP_DROP_SUBCORPUS('%s')\n", subcorpus);
 
   /* make sure it is a subcorpus, not a root corpus */
   if (!split_subcorpus_spec(subcorpus, &c, &sc))
@@ -1161,7 +1261,8 @@ do_cqi_cqp_drop_subcorpus(void) {
 
 /* temporary functions for CQI_CQP_FDIST_1() and CQI_CQP_FDIST_2() */
 void
-do_cqi_cqp_fdist_1(void) {
+do_cqi_cqp_fdist_1(void)
+{
   char *subcorpus;
   CorpusList *cl;
   int cutoff;
@@ -1188,7 +1289,7 @@ do_cqi_cqp_fdist_1(void) {
     fieldtype = field_name_to_type(fieldname);
   }
   if (server_debug) 
-    fprintf(stderr, "CQi: CQI_CQP_FDIST_1('%s', %d, %s, %s)\n", 
+    Rprintf("CQi: CQI_CQP_FDIST_1('%s', %d, %s, %s)\n", 
             subcorpus, cutoff, fieldname, att);
   
   cl = cqi_find_corpus(subcorpus);
@@ -1199,7 +1300,7 @@ do_cqi_cqp_fdist_1(void) {
   else {
     /* compute_grouping() returns tokens with f > cutoff, but CQi specifies f >= cutoff */
     cutoff = (cutoff > 0) ? cutoff - 1 : 0;
-    table = compute_grouping(cl, NoField, 0, NULL, fieldtype, 0, att, cutoff);
+    table = compute_grouping(cl, NoField, 0, NULL, fieldtype, 0, att, cutoff, 0);
     if (table == NULL) {
       cqi_command(CQI_CQP_ERROR_GENERAL);
     }
@@ -1223,7 +1324,8 @@ do_cqi_cqp_fdist_1(void) {
 
 
 void
-do_cqi_cqp_fdist_2(void) {
+do_cqi_cqp_fdist_2(void)
+{
   char *subcorpus;
   CorpusList *cl;
   int cutoff;
@@ -1260,7 +1362,7 @@ do_cqi_cqp_fdist_2(void) {
     fieldtype2 = field_name_to_type(fieldname2);
   }
   if (server_debug) 
-    fprintf(stderr, "CQi: CQI_CQP_FDIST_2('%s', %d, %s, %s, %s, %s)\n", 
+    Rprintf("CQi: CQI_CQP_FDIST_2('%s', %d, %s, %s, %s, %s)\n", 
             subcorpus, cutoff, fieldname1, att1, fieldname2, att2);
   
   cl = cqi_find_corpus(subcorpus);
@@ -1271,7 +1373,7 @@ do_cqi_cqp_fdist_2(void) {
   else {
     /* compute_grouping() returns tokens with f > cutoff, but CQi specifies f >= cutoff */
     cutoff = (cutoff > 0) ? cutoff - 1 : 0;
-    table = compute_grouping(cl, fieldtype1, 0, att1, fieldtype2, 0, att2, cutoff);
+    table = compute_grouping(cl, fieldtype1, 0, att1, fieldtype2, 0, att2, cutoff, 0);
     if (table == NULL) {
       cqi_command(CQI_CQP_ERROR_GENERAL);
     }
@@ -1296,14 +1398,18 @@ do_cqi_cqp_fdist_2(void) {
 }
 
 
-/*
+/**
  *
- *  The command interpreter loop (returns on exit)
+ *  The CQP server's command interpreter loop.
+ *
+ *  The loops starts running when this function is called, and when the
+ *  exit command is reveived (CQI_CTRL_BYE)
+ *  (returns on exit)
  *
  */
-
 void 
-interpreter(void) {
+interpreter(void)
+{
   int cmd;
   int cmd_group;
 
@@ -1317,28 +1423,29 @@ interpreter(void) {
     case CQI_CTRL:
       switch (cmd) {
       case CQI_CTRL_CONNECT:
-        wrong_command_error(cmd);
+        cqiserver_wrong_command_error(cmd);
+        break;
       case CQI_CTRL_BYE:
         if (server_debug) 
-          fprintf(stderr, "CQi: CQI_CTRL_BYE()\n");
+          Rprintf("CQi: CQI_CTRL_BYE()\n");
         cqi_command(CQI_STATUS_BYE_OK);
         return;                 /* exit CQi command interpreter */
       case CQI_CTRL_USER_ABORT:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_CTRL_ABORT signal ... ignored\n");
+          Rprintf("CQi: CQI_CTRL_ABORT signal ... ignored\n");
         break;
       case CQI_CTRL_PING:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_CTRL_PING()\n");
+          Rprintf("CQi: CQI_CTRL_PING()\n");
         cqi_command(CQI_STATUS_PING_OK);
         break;
       case CQI_CTRL_LAST_GENERAL_ERROR:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_CTRL_LAST_GENERAL_ERROR() => '%s'", cqi_error_string);
+          Rprintf("CQi: CQI_CTRL_LAST_GENERAL_ERROR() => '%s'", cqi_error_string);
         cqi_data_string(cqi_error_string);
         break;
       default:
-        unknown_command_error(cmd);
+        cqiserver_unknown_command_error(cmd);
       }
       break;
       
@@ -1347,22 +1454,22 @@ interpreter(void) {
       switch (cmd) {
       case CQI_ASK_FEATURE_CQI_1_0:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_ASK_FEATURE_CQI_1_0 ... CQi v1.0 ok\n");
+          Rprintf("CQi: CQI_ASK_FEATURE_CQI_1_0 ... CQi v1.0 ok\n");
         cqi_data_bool(CQI_CONST_YES);
         break;
       case CQI_ASK_FEATURE_CL_2_3:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_ASK_FEATURE_CL_2_3 ... CL v2.3 ok\n");
+          Rprintf("CQi: CQI_ASK_FEATURE_CL_2_3 ... CL v2.3 ok\n");
         cqi_data_bool(CQI_CONST_YES);
         break;
       case CQI_ASK_FEATURE_CQP_2_3:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_ASK_FEATURE_CQP_2_3 ... CQP v2.3 ok\n");
+          Rprintf("CQi: CQI_ASK_FEATURE_CQP_2_3 ... CQP v2.3 ok\n");
         cqi_data_bool(CQI_CONST_YES);
         break;
       default:
         if (server_debug)
-          fprintf(stderr, "CQi: CQI_ASK_FEATURE_* ... <unknown feature> not supported\n");
+          Rprintf("CQi: CQI_ASK_FEATURE_* ... <unknown feature> not supported\n");
         cqi_data_bool(CQI_CONST_NO);
       }
       break;
@@ -1395,7 +1502,7 @@ interpreter(void) {
         do_cqi_corpus_full_name();
         break;
       default:
-        unknown_command_error(cmd);
+        cqiserver_unknown_command_error(cmd);
       }
       break;
       
@@ -1457,7 +1564,7 @@ interpreter(void) {
         do_cqi_cl_alg2cpos();
         break;
       default:
-        unknown_command_error(cmd);
+        cqiserver_unknown_command_error(cmd);
       }
       break;
 
@@ -1489,33 +1596,60 @@ interpreter(void) {
         do_cqi_cqp_fdist_2();
         break;
       default:
-        unknown_command_error(cmd);
+        cqiserver_unknown_command_error(cmd);
       }
       break;
       
     default:
-      unknown_command_error(cmd);
-    }
+      cqiserver_unknown_command_error(cmd);
+
+    } /* end outer switch */
     
-  }
+  } /* end while 42 */
 
 }
 
 
 
-
-int main(int argc, char *argv[]) {
+/**
+ * Main function for the cqpserver app.
+ */
+int
+main(int argc, char *argv[])
+{
   int cmd;
 
   which_app = cqpserver;
 
+  /* TODO: shouldn't these come AFTER initialize_cqp(), as that function may overwrite these values with defaults?
+   * or maybe I've missed some subtlety here....*/
   silent = 1; 
   paging = autoshow = auto_save = 0;
 
   if (!initialize_cqp(argc, argv)) {
-    fprintf(stderr, "ERROR Couldn't initialise CQP engine.\n");
+    Rprintf("CQPserver: ERROR Couldn't initialise CQP engine.\n");
     exit(1);
   }
+  while (optind < argc) {
+    /* remaining command-line arguments are <user>:<password> specifications */
+    char *sep = strchr(argv[optind], ':');
+    if (sep != NULL) {
+      if (sep == argv[optind]) {
+        Rprintf("CQPserver: Invalid account specification '%s' (username must not be empty)\n", argv[optind]);
+        exit(1);
+      }
+      else {
+        *sep = '\0';
+        add_user_to_list(argv[optind], sep + 1);
+      }
+    }
+    else {
+      Rprintf("CQPserver: Invalid account specification '%s' (password missing)\n", argv[optind]);
+      exit(1);
+    }
+    optind++;
+  }
+  
   cqiserver_welcome();
 
   if (localhost) {
@@ -1524,10 +1658,10 @@ int main(int argc, char *argv[]) {
 
   if (0 < accept_connection(server_port)) {
     if (server_log)
-      printf("Connected. Waiting for CONNECT request.\n");
+      Rprintf("CQPserver: Connected. Waiting for CONNECT request.\n");
   }
   else {
-    fprintf(stderr, "ERROR Connection failed.\n");
+    Rprintf("CQPserver: ERROR Connection failed.\n");
     exit(1);
   }
 
@@ -1535,17 +1669,17 @@ int main(int argc, char *argv[]) {
   cmd = cqi_read_command();
   if (cmd != CQI_CTRL_CONNECT) {
     if (server_log)
-      printf("Connection refused.\n");
-    wrong_command_error(cmd);
+      Rprintf("CQPserver: Connection refused.\n");
+    cqiserver_wrong_command_error(cmd);
   }
   user = cqi_read_string();
   passwd = cqi_read_string();
   if (server_log)
-    printf("CONNECT  user = '%s'  passwd = '%s'  pid = %d\n", user, passwd, (int)getpid());
+    Rprintf("CQPserver: CONNECT  user = '%s'  passwd = '%s'  pid = %d\n", user, passwd, (int)getpid());
 
   /* check password here (always required !!) */
   if (!authenticate_user(user, passwd)) {
-    printf("Wrong password. Connection refused.\n");
+    Rprintf("CQPserver: Wrong username or password. Connection refused.\n"); /* TODO shouldn't this be to stderr as it is not conditional on server_log? */
     cqi_command(CQI_ERROR_CONNECT_REFUSED);
   }
   else {
@@ -1556,24 +1690,25 @@ int main(int argc, char *argv[]) {
 
     /* check which corpora the user is granted access to */
     {
-    CorpusList *cl = FirstCorpusFromList();
-    while (cl != NULL) {
-      if (!check_grant(user, cl->name))
-        dropcorpus(cl);
-      cl = NextCorpusFromList(cl);
-    }
+      CorpusList *cl = FirstCorpusFromList();
+      while (cl != NULL) {
+        if (!check_grant(user, cl->name))
+          dropcorpus(cl);
+        cl = NextCorpusFromList(cl);
+      }
     }
 
     /* start command interpreter loop */
     interpreter();
 
     if (server_log)
-      printf("User '%s' has logged off from the CQPserver.\n", user);
+      Rprintf("CQPserver: User '%s' has logged off.\n", user);
   }
 
   /* connection terminated; clean up and exit */
-  printf("Exit. (pid = %d)\n", (int)getpid());
+  Rprintf("CQPserver: Exit. (pid = %d)\n", (int)getpid());
 
+  /* TODO should we check cqp_error_status as in the main cqp app? */
   return 0;
 }
 
